# Pruning by Active Attention Manipulation

## Requirements

The code is tested on `python==3.9, pyhocon==0.3.57, torch=1.8.0, torchvision=0.9.0`

## Get Started

You can start pruning a Cifar ResNet with following command, entering the path to the model trained to warm-up epochs

**Single node, single GPU:**

```bash
python3 codebase/res_prune.py --model ResNet56 --landa 0.0005 --input_path warm-up_model_dir --output_path Res56Pruned
```


You can start pruning a ResNet50 with following command:

**Single node, single GPU:**

```bash
CUDA_VISIBLE_DEVICES=0 python3 -m entry.prune50 --conf conf/resnet50-tfrec-v1_5.conf -o output/res50/lambda0001 model.pretrained=true -M data.use_dali=False data.root=root_path only_evaluate=false
```

## Warm-Up

You can train the Cifar models to desired warm-up epochs with the following command. 

```bash
python3 -m entry.run --conf conf/cifar10.conf -o output/res56/warmup -M model.name=ResNet56_vanilla max_epochs=50 only_evaluate=false 
```

## Fine-Tuning

You can fine-tune the pruned model with the following command, entering the model name. Enter the output path of the pruned model as the input directory (-i). 

```bash
python3 -m entry.run --conf conf/cifar10.conf -i Res56Pruned -o output/res56/tuned only_evaluate=false -M model.name=ResNet56
```


## Features

This codebase adopt configuration file (`.hocon`) to store the hyperparameters (such as the learning rate, training epochs and etc.).
If you want to modify the configuration hyperparameters, you have two ways:

1. Modify the configuration file to generate a new file.

2. You can add `-M` in the running command line to modify the hyperparameters temporarily.


For example, if you hope to modify the total tuning epochs to 100 and the learning rate to 0.05. You can run the following command:

```bash
python3 -m entry.run --conf conf/cifar10.conf -i Res56Pruned -o output/res56/tuned -M max_epochs=100 optimizer.lr=0.05
```

If you modify a non existing hyperparameter, the code will raise an exception.

To list all valid hyperparameters names, you can run the following command:

```bash
pyhocon -i conf/cifar10.conf -f properties
```

